'use strict';

var StoreError = function(message) {
  this.message = message;
};
var StoreNotFoundError = function(message) {
  this.message = message;
};
StoreNotFoundError.prototype = Object.create(StoreError);


var Store = function(parent, scopeFilters) {
  this.parent = parent;
  this.scopeFilters = scopeFilters;
  this._clear(true);
};

Store.prototype.indexProperties = [];
Store.prototype.typeNameProperty = '__type__';
Store.prototype.typeName = null;
Store.prototype.dropDuplicates = false;

Store.prototype._assertNoParent = function() {
  if (this.parent) {
    throw new StoreError('This method cannot be called from scoped store');
  }
};

Store.prototype.root = function() {
  return this.parent ? this.parent : this;
};

Store.prototype._ensureIsParent = function(cb) {
  return cb.call(this.root());
};

Store.prototype.format = function(obj) {
  return obj;
};

Store.prototype.match = function(obj) {
  if (this.scopeFilters) {
    return StoreQuery.matchObject(obj, this.scopeFilters);
  }
  return true;
};

Store.prototype.exists = function(id) {
  var data = this.parent ? this.parent.data : this.data;
  if (!data.exists(id)) {
    return false;
  }
  return this.match(data.get(id));
};

Store.prototype.get = function(id) {
  if (!this.exists(id)) {
    throw new StoreNotFoundError('Object "' + id + '" not found' +
      (this.typeName ? (' in ' + this.typeName) : ''));
  }
  return (this.parent ? this.parent.data : this.data).get(id);
};

Store.prototype.tryGet = function(id) {
  try {
    return this.get(id);
  } catch (e) {
    if (e instanceof StoreNotFoundError) {
      return null;
    }
    throw e;
  }
};

Store.prototype.createQuery = function(filters, processor) {
  if (this.scopeFilters) {
    filters = StoreQuery.mergeFilters(this.scopeFilters, filters);
  }
  var data = this.parent ? this.parent.data : this.data;
  return new StoreQuery(data.all(), filters, processor, data.getIndexes());
};

Store.prototype.query = function(filters, processor) {
  return this.createQuery(filters, processor).execute();
};

Store.prototype.count = function() {
  if (this.scopeFilters) {
    return this.query().length;
  }
  return (this.parent ? this.parent.data : this.data).count();
};

Store.prototype.insert = function(obj) {
  return this._ensureIsParent(function() {
    obj = this.format(obj);
    if (this.typeName) {
      obj[this.typeNameProperty] = this.typeName;
    }
    if (obj.id && this.exists(obj.id)) {
      if (this.dropDuplicates) {
        this.data.remove(this.data.get(obj.id));
      } else {
        throw new StoreError('Duplicate id "' + obj.id + '"' +
          (this.typeName ? (' in ' + this.typeName) : ''));
      }
    }
    this.data.push(obj);
    this._afterChange('inserted', obj);
    return obj;
  });
};

Store.prototype.update = function(obj, props, updateSpecifier) {
  return this._ensureIsParent(function() {
    obj = angular.extend(this.data.get(obj.id), obj);

    if (props) {
      angular.forEach(props, function(value, key) {
        if (typeof(value) === 'function') {
          obj[key] = value(obj[key], obj);
        } else {
          obj[key] = value;
        }
      });
    }

    obj = this.format(obj);
    this.data._index(obj);
    this._afterChange('updated', obj, updateSpecifier);
    return obj;
  });
};

Store.prototype.updateById = function(id, data, updateSpecifier) {
  return this._ensureIsParent(function() {
    var obj = this.data.get(id);
    if (typeof(data.id) !== 'undefined' && data.id !== id) {
      obj = this.data.update({id: data.id}, id);
    }
    return this.update(obj, data, updateSpecifier);
  });
};

Store.prototype.upsert = function(obj, props, updateSpecifier) {
  return this._ensureIsParent(function() {
    obj = this.format(obj);
    if (!this.data.exists(obj.id)) {
      angular.extend(obj, props);
      return this.insert(obj);
    }
    return this.update(obj, props, updateSpecifier);
  });
};

Store.prototype.remove = function(obj) {
  var args = Array.prototype.slice.call(Array.from(arguments), 1);
  return this._ensureIsParent(function() {
    obj = this.data.remove(obj);
    this._afterChange.apply(this, ['removed', obj].concat(args));
    return obj;
  });
};

Store.prototype.removeById = function(id) {
  this.remove(this.get(id));
};

Store.prototype._createData = function() {
  return new StoreData(this.indexProperties);
};

Store.prototype._clear = function(clearListeners) {
  if (!this.parent) {
    this.data = this._createData();
  }
  this.locks = {};
  if (clearListeners) {
    this.listeners = {};
  }
};

Store.prototype.clear = function(clearListeners) {
  return this._ensureIsParent(function() {
    this._clear(clearListeners);
    this._afterChange('cleared');
    return this;
  });
};

var _StoreManyGenerator = function(method) {
  return function(objects) {
    if (!objects || !objects.length) {
      return [];
    }
    var args = Array.prototype.slice.call(Array.from(arguments), 1);
    return this._ensureIsParent(function() {
      var changed = [];
      this.transaction(function() {
        objects.forEach(function(obj) {
          changed.push(this[method].apply(this, [obj].concat(args)));
        }, this);
      });
      return changed;
    });
  };
};

Store.prototype.insertMany = _StoreManyGenerator('insert');

Store.prototype.updateMany = _StoreManyGenerator('update');

Store.prototype.upsertMany = _StoreManyGenerator('upsert');

Store.prototype.removeMany = function(objects) {
  if (!objects || !objects.length) {
    return [];
  }
  var args = Array.prototype.slice.call(Array.from(arguments), 1);
  return this._ensureIsParent(function() {
    var removed = [];
    this.transaction(function() {
      for (var i = objects.length - 1; i >= 0; i--) {
        removed.push(this.remove.apply(this, [objects[i]].concat(args)));
      }
    });
    return removed;
  });
};

Store.prototype.removeAll = function() {
  var args = Array.from(arguments);
  return this.removeMany(this.query.apply(this, args));
};

Store.prototype.removeAllNotMatching = function(objects) {
  var ids = objects.map(function(o) { return o.id; });
  var q = {id: function(id) {
    return ids.indexOf(id) === -1;
  }};
  return this.removeMany(this.query(q));
};

Store.prototype.replace = function(objects) {
  this.transaction(function() {
    this.removeAll();
    this.upsertMany(objects);
  }, this);
};

Store.prototype.sync = function(objects) {
  return this._ensureIsParent(function() {
    this.transaction(function() {
      var actualIds = [];
      objects.forEach(function(obj) {
        this.upsert(obj);
        actualIds.push(obj.id);
      }, this);
      this.data.ids().forEach(function(id) {
        if (actualIds.indexOf(id) === -1) {
          this.remove(this.data.get(id));
        }
      }, this);
    });
    return this;
  });
};

Store.prototype.transaction = function(callback, thisObject) {
  return this._ensureIsParent(function() {
    if (this._delayStoreUpdate) {
      return callback.call(thisObject || this);
    }
    this._delayStoreUpdate = true;
    this._delayedStoreUpdates = [];
    var r = callback.call(thisObject || this);
    this._afterChangeCallback();
    this._delayStoreUpdate = false;

    var broadcastedEvents = {};
    this._delayedStoreUpdates.forEach(function(args) {
      if (args[1]) {
        if (typeof(broadcastedEvents[args[1].id]) === 'undefined') {
          broadcastedEvents[args[1].id] = [];
        }
        if (broadcastedEvents[args[1].id].indexOf(args[0]) !== -1) {
          return;
        }
        broadcastedEvents[args[1].id].push(args[0]);
      }
      this.broadcast.apply(this, args);
    }, this);
    this._delayedStoreUpdates = null;

    this.broadcast('changed');
    return r;
  });
};

Store.prototype._afterChange = function(eventName) {
  var args = Array.from(arguments);
  if (this._delayStoreUpdate) {
    this._delayedStoreUpdates.push(args);
    return;
  }
  this._afterChangeCallback();
  this.broadcast.apply(this, args);
  this.broadcast('changed');
};

Store.prototype._afterChangeCallback = function() {
  // custom logic
};

Store.prototype._getListeners = function(eventName) {
  if (typeof(this.listeners[eventName]) === 'undefined') {
    this.listeners[eventName] = [];
  }
  return this.listeners[eventName];
};

Store.prototype.broadcast = function(eventName) {
  var args = Array.prototype.slice.call(Array.from(arguments), 1);
  // copy the array before iterating as listeners can be modified on the go
  this._getListeners(eventName).slice(0).forEach(function(listener) {
    listener.apply(this, args);
  }, this);
};

Store.prototype.on = function(eventName, callback) {
  if (this.parent) {
    var self = this;
    return this.parent.on(eventName, function(obj) {
      if (obj && !self.match(obj)) {
        return;
      }
      var args = Array.from(arguments);
      callback.apply(self, args);
    });
  }

  // https://stackoverflow.com/a/12996879
  if (Object.prototype.toString.call(eventName) === '[object Array]') {
    return this._offMany(eventName.map(function(e) {
      return this.on(e, callback);
    }, this));
  }
  var listeners = this._getListeners(eventName);
  listeners.push(callback);
  var off = function() {
    var i = listeners.indexOf(callback);
    if (i >= 0) {
      listeners.splice(i, 1);
    }
  };
  off.callback = callback;
  return off;
};

Store.prototype.watch = function(obj, eventNames, callback, immediate) {
  if (typeof(eventNames) === 'function') {
    immediate = callback;
    callback = eventNames;
    eventNames = null;
  } else if (typeof(eventNames) === 'string') {
    eventNames = [eventNames];
  }
  if (!eventNames) {
    eventNames = ['inserted', 'updated', 'removed'];
  }

  if (typeof(obj) !== 'object') {
    obj = {id: obj};
  }

  if (this.parent) {
    var self = this;
    return this.parent.watch(obj, eventNames, function(obj) {
      if (!self.match(obj)) {
        return;
      }
      callback.call(self, obj);
    }, immediate);
  }

  var off = this._offMany(eventNames.map(function(e) {
    return this.on(e, function(o) {
      if (o.id === obj.id) {
        callback.apply(this, [].concat(Array.from(arguments), [e]));
      }
    });
  }, this));

  if ((immediate === undefined || immediate) && this.exists(obj.id)) {
    callback.call(this, this.get(obj.id));
  }
  return off;
};

Store.prototype.watchAll = function(callback, immediate) {
  var off = this.on('changed', callback);
  if (callback && (immediate === undefined || immediate)) {
    callback.call(this);
  }
  return off;
};

Store.prototype._offMany = function(offs) {
  return function() {
    offs.forEach(function(off) { off && off(); });
  };
};

Store.prototype._lock = function(lockId, callback, executor) {
  var lock = this._getLock(lockId);
  if (lock) {
    lock.wait(callback);
    return lock;
  }

  var self = this;
  this.locks[lockId] = lock = {
    id: lockId,
    released: false,
    callbacks: [],
    releaseArgs: [],
    wait: function(cb) {
      if (this.released) {
        cb.apply(self, this.releaseArgs);
      } else {
        this.callbacks.push(cb);
      }
    },
    release: function(dropCallbacks) {
      delete self.locks[lockId];
      this.released = true;
      if (!dropCallbacks) {
        var args = this.releaseArgs.concat(Array.prototype.slice.call(Array.from(arguments), 1));
        this.callbacks.forEach(function(cb) {
          if (cb) {
            cb.apply(self, args);
          }
        });
      }
    },
    success: function() {
      this.release.apply(this, [false].concat(Array.from(arguments)));
    },
    error: function() {
      this.release(true);
    },
    wrap: function(cb) {
      try {
        this.release(false, cb());
      } catch(e) {
        this.release(true);
      }
    }
  };

  if (callback) {
    lock.callbacks.push(callback);
  }
  if (executor) {
    executor(lock.success.bind(lock), lock.error.bind(lock), lock);
  }
  return lock;
};

Store.prototype._getLock = function(lockId) {
  if (typeof(this.locks[lockId]) !== 'undefined') {
    return this.locks[lockId];
  }
  return null;
};

Store.prototype.lock = function(obj, callback, executor) {
  var self = this;
  var lock = this._lock(obj.id, function() {
    self.broadcast('unlocked', obj);
    if (callback) {
      callback.apply(self, arguments);
    }
  }, executor);
  lock.releaseArgs = [obj];
  this.broadcast('locked', obj);
  return lock;
};

Store.prototype.isLocked = function(obj) {
  return this._getLock(obj.id) !== null;
};

Store.prototype.unlocked = function(obj, callback) {
  if (typeof(this.locks[obj.id]) !== 'undefined') {
    this.locks[obj.id].wait(callback);
  } else {
    callback(obj);
  }
};

Store.prototype.scoped = function(filters) {
  if (this.parent) {
    return this.parent.scoped(StoreQuery.mergeFilters(this.scopeFilters || {}, filters));
  }
  return new this.constructor(this, filters);
};

Store.prototype.join = function(prop, store, targetProp) {
  var filter = {};
  filter[prop] = StoreQuery.any(store.query().pluck(targetProp || 'id'));
  return this.scoped(filter);
};

Store.computeMoveBounds = function(currentPosition, position) {
  if (currentPosition === null) {
    return {'shift': 1, 'lower': position, 'upper': null};
  } else if (position === null) {
    return {'shift': -1, 'lower': currentPosition, 'upper': null};
  }
  var up = position > currentPosition;
  var shift = up ? -1 : 1;
  var lower = Math.min(currentPosition, position);
  lower += up ? 1 : 0;
  var upper = Math.max(currentPosition, position);
  upper -= up ? 0 : 1;
    return {'shift': shift, 'lower': lower, 'upper': upper};
};
